/*
Velleman K8055 USB Experiment Interface Board library for BeOS/ZETA
Copyright (c) 2007 Gabriele Biffi
Version Beta 1 - 2007/02/09

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <stdio.h>

#include <Alert.h>
#include <Application.h>
#include <Box.h>
#include <Button.h>
#include <CheckBox.h>
#include <MenuItem.h>
#include <MenuField.h>
#include <Screen.h>
#include <Slider.h>
#include <StringView.h>
#include <View.h>
#include <Window.h>

#include "k8055.h"

// Messages

#define MSG_CONNECT				'Conn'
#define MSG_SET_ALL_DIGITAL		'SADi'
#define MSG_CLEAR_ALL_DIGITAL	'CADi'
#define MSG_SET_ALL_ANALOG		'SAAn'
#define MSG_CLEAR_ALL_ANALOG	'CAAn'
#define MSG_OUTPUT_TEST			'OutT'
#define MSG_SET_OUTPUTS			'OutS'
#define MSG_RESET_COUNTER		'CRes'
#define MSG_SET_DEBOUNCE_TIME	'CDeb'

// Classes declarations ------------------------------------------------------

class MyWindow;

class MyView : public BView
{
private:
			MyWindow *	fParent;
public:
						MyView(MyWindow *window);
	virtual	void		Pulse(void);
};

class CounterBox : public BBox
{
public:
					CounterBox(BRect frame, int counterNumber);
	BStringView *	fCounterValue;
};

class MyWindow : public BWindow
{
friend class MyView;
private:
	// Properties
			int				fTest;
			bool			fConnected;
	// Board select and connect
			BCheckBox *		fSK5;
			BCheckBox *		fSK6;
			BButton *		fConnect;
			BStringView *	fConnectLabel;
	// Analogs
			BSlider *		fDAC1;
			BStringView *	fDAC1Label;
			BSlider *		fDAC2;
			BStringView *	fDAC2Label;
			BSlider *		fADC1;
			BStringView *	fADC1Label;
			BSlider *		fADC2;
			BStringView *	fADC2Label;
	// Digital
			BCheckBox *		fDigitalIn[5];
			BCheckBox *		fDigitalOut[8];
	// Counters
			CounterBox *	fCounter1;
			CounterBox *	fCounter2;
	// Methods
			void			_update_analog_labels(void);
public:
							MyWindow(void);
	virtual	void			MessageReceived(BMessage *message);
	virtual	bool			QuitRequested(void);
};

class MyApp : public BApplication
{
private:
			MyWindow *	fWindow;
public:
						MyApp(void);
	virtual	void		AboutRequested(void);
	virtual	void		ReadyToRun(void);
};

// MyView --------------------------------------------------------------------

MyView::MyView(MyWindow *window) :
	BView(window->Bounds(), "background_view", B_FOLLOW_ALL_SIDES,
		B_PULSE_NEEDED)
{
	fParent = window;
	SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));
}

void MyView::Pulse(void)
{
	if (!fParent->fConnected) return;

	int analog1 = 0, analog2 = 0, digital = 0, counter, i;
	char buffer[32];

	// Test in progress?
	if (fParent->fTest > -1) {
		fParent->fTest++;
		if (fParent->fTest >= 11)
			fParent->fTest = 1;
		if (fParent->fTest >= 1 && fParent->fTest <= 8)
			digital = 0x01 << (fParent->fTest - 1);
		if (fParent->fTest == 9)
			analog1 = 255;
		if (fParent->fTest == 10)
			analog2 = 255;
		WriteAllDigital(digital);
		OutputAllAnalog(analog1, analog2);
		for (i = 0; i < 8; i++)
			fParent->fDigitalOut[i]->SetValue(digital & (0x01 << i) ?
				B_CONTROL_ON : B_CONTROL_OFF);
		fParent->fDAC1->SetValue(analog1);
		fParent->fDAC2->SetValue(analog2);
		fParent->_update_analog_labels();
	}

	// Update status
	ReadAllAnalog(&analog1, &analog2);
	if (analog1 >= 0 && analog1 <= 255) {
		sprintf(buffer, "%d", analog1);
		fParent->fADC1Label->SetText(buffer);
		fParent->fADC1->SetValue(analog1);
	}
	if (analog2 >= 0 && analog2 <= 255) {
		sprintf(buffer, "%d", analog2);
		fParent->fADC2Label->SetText(buffer);
		fParent->fADC2->SetValue(analog2);
	}
	digital = ReadAllDigital();
	if (digital >= 0) {
		for (i = 0; i < 5; i++)
			fParent->fDigitalIn[i]->SetValue(digital & (0x01 << i) ?
				B_CONTROL_ON : B_CONTROL_OFF);
	}
	counter = ReadCounter(1);
	if (counter > -1) {
		sprintf(buffer, "%d", counter);
		fParent->fCounter1->fCounterValue->SetText(buffer);
	}
	counter = ReadCounter(2);
	if (counter > -1) {
		sprintf(buffer, "%d", counter);
		fParent->fCounter2->fCounterValue->SetText(buffer);
	}
}

// CounterBox ----------------------------------------------------------------

CounterBox::CounterBox(BRect frame, int counterNumber) :
	BBox(frame, NULL)
{
	const int kDebounceTimes[4] = {0, 2, 10, 1000};
	BMessage *message;
	BButton *button;
	char buffer[32];
	BMenu *menu;
	BRect r;

	sprintf(buffer, "counter_%d_box", counterNumber);
	SetName(buffer);

	sprintf(buffer, "Counter %d", counterNumber);
	SetLabel(buffer);

	// Counter value
	r = BRect(5, 15, Bounds().right - 5, 30);
	sprintf(buffer, "counter_%d_value", counterNumber);
	fCounterValue = new BStringView(r, buffer, "0",
		B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP);
	fCounterValue->SetAlignment(B_ALIGN_CENTER);
	AddChild(fCounterValue);

	// Reset button
	message = new BMessage(MSG_RESET_COUNTER);
	message->AddInt32("counter_number", counterNumber);
	r.top = 35;
	r.bottom = 55;
	sprintf(buffer, "counter_%d_reset", counterNumber);
	button = new BButton(r, buffer, "Reset", message,
		B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP);
	AddChild(button);

	// Debounce time
	menu = new BMenu("Debounce time");
	for (int i = 0; i < 4; i++) {
		message = new BMessage(MSG_SET_DEBOUNCE_TIME);
		message->AddInt32("counter_number", counterNumber);
		message->AddInt32("debounce_time", kDebounceTimes[i]);
		sprintf(buffer, "%d ms", kDebounceTimes[i]);
		menu->AddItem(new BMenuItem(buffer, message));
	}
	r.top = 65;
	r.bottom = 85;
	sprintf(buffer, "counter_%d_debounce", counterNumber);
	AddChild(new BMenuField(r, buffer, NULL, menu, true,
		B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP));

	ResizeTo(Bounds().Width(), 90);
}

// MyWindow ------------------------------------------------------------------

MyWindow::MyWindow(void) :
	BWindow(BRect(0, 0, 565, 375), "K8055 Demo", B_TITLED_WINDOW,
		B_ASYNCHRONOUS_CONTROLS | B_NOT_RESIZABLE | B_NOT_ZOOMABLE)
{
	BBox *boardBox, *analogBox, *inBox, *outBox;
	char buffer1[32], buffer2[32];
	BView *back, *buttonsBox;
	BStringView *label;
	BButton *button;
	BScreen screen;
	int i;

	// We are not connected at start
	fConnected = false;

	// We are not performing output test
	fTest = -1;

	// Set pulse rate for status update (5 updates/sec)
	SetPulseRate(200000);

	// Background view. Also serves as status updater.
	back = new MyView(this);
	AddChild(back);

	// Board select and connect
	boardBox = new BBox(BRect(5, 5, 130, 95), "board_box");
	boardBox->SetLabel("Board");
	fSK5 = new BCheckBox(BRect(5, 15, 60, 35), "sk5_check", "SK5", NULL);
	fSK5->SetValue(B_CONTROL_ON);
	boardBox->AddChild(fSK5);
	fSK6 = new BCheckBox(BRect(65, 15, 120, 35), "sk6_check", "SK6", NULL);
	fSK6->SetValue(B_CONTROL_ON);
	boardBox->AddChild(fSK6);
	fConnect = new BButton(BRect(5, 40, 120, 65), "connect_button", "Connect",
		new BMessage(MSG_CONNECT));
	boardBox->AddChild(fConnect);
	fConnectLabel = new BStringView(BRect(5, 70, 120, 85), "connect_label",
		"...");
	boardBox->AddChild(fConnectLabel);
	back->AddChild(boardBox);

	// Buttons
	buttonsBox = new BView(BRect(10, 160, 125, 365), "buttons_box",
		B_FOLLOW_LEFT | B_FOLLOW_TOP, 0);
	buttonsBox->SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));
	button = new BButton(BRect(0, 0, 115, 20), "set_all_digital_button",
		"Set all digital", new BMessage(MSG_SET_ALL_DIGITAL));
	buttonsBox->AddChild(button);
	button = new BButton(BRect(0, 25, 115, 45), "clear_all_digital_button",
		"Clear all digital", new BMessage(MSG_CLEAR_ALL_DIGITAL));
	buttonsBox->AddChild(button);
	button = new BButton(BRect(0, 60, 115, 80), "set_all_analog_button",
		"Set all analog", new BMessage(MSG_SET_ALL_ANALOG));
	buttonsBox->AddChild(button);
	button = new BButton(BRect(0, 85, 115, 105), "clear_all_analog_button",
		"Clear all analog", new BMessage(MSG_CLEAR_ALL_ANALOG));
	buttonsBox->AddChild(button);
	button = new BButton(BRect(0, 120, 115, 140), "output_test_button",
		"Output test", new BMessage(MSG_OUTPUT_TEST));
	buttonsBox->AddChild(button);
	button = new BButton(BRect(0, 155, 115, 175), "about_button",
		"About...", new BMessage(B_ABOUT_REQUESTED));
	button->SetTarget(be_app_messenger);
	buttonsBox->AddChild(button);
	button = new BButton(BRect(0, 180, 115, 200), "close_button", "Close",
		new BMessage(B_QUIT_REQUESTED));
	buttonsBox->AddChild(button);
	back->AddChild(buttonsBox);

	// Analogs
	analogBox = new BBox(BRect(135, 5, 300, 365), "analog_box");
	analogBox->SetLabel("Analogs");
	// DAC1
	label = new BStringView(BRect(5, 15, 40, 30), "dac1", "DAC1");
	label->SetAlignment(B_ALIGN_CENTER);
	analogBox->AddChild(label);
	fDAC1 = new BSlider(BRect(5, 35, 40, 335), "dac1_slider", NULL,
		new BMessage(MSG_SET_OUTPUTS), 0, 255, B_BLOCK_THUMB,
		B_FOLLOW_LEFT | B_FOLLOW_TOP_BOTTOM);
	fDAC1->SetModificationMessage(new BMessage(MSG_SET_OUTPUTS));
	fDAC1->SetOrientation(B_VERTICAL);
	analogBox->AddChild(fDAC1);
	fDAC1Label = new BStringView(BRect(5, 340, 40, 355), "dac1_label", "0",
		B_FOLLOW_LEFT | B_FOLLOW_BOTTOM);
	fDAC1Label->SetAlignment(B_ALIGN_CENTER);
	analogBox->AddChild(fDAC1Label);
	// DAC2
	label = new BStringView(BRect(45, 15, 80, 30), "dac2", "DAC2");
	label->SetAlignment(B_ALIGN_CENTER);
	analogBox->AddChild(label);
	fDAC2 = new BSlider(BRect(45, 35, 80, 335), "dac2_slider", NULL,
		new BMessage(MSG_SET_OUTPUTS), 0, 255, B_BLOCK_THUMB,
		B_FOLLOW_LEFT | B_FOLLOW_TOP_BOTTOM);
	fDAC2->SetModificationMessage(new BMessage(MSG_SET_OUTPUTS));
	fDAC2->SetOrientation(B_VERTICAL);
	analogBox->AddChild(fDAC2);
	fDAC2Label = new BStringView(BRect(45, 340, 80, 355), "dac2_label", "0",
		B_FOLLOW_LEFT | B_FOLLOW_BOTTOM);
	fDAC2Label->SetAlignment(B_ALIGN_CENTER);
	analogBox->AddChild(fDAC2Label);
	// ADC1
	label = new BStringView(BRect(85, 15, 120, 30), "adc1", "ADC1");
	label->SetAlignment(B_ALIGN_CENTER);
	analogBox->AddChild(label);
	fADC1 = new BSlider(BRect(85, 35, 120, 335), "adc1_slider", NULL,
		new BMessage(MSG_SET_OUTPUTS), 0, 255, B_BLOCK_THUMB,
		B_FOLLOW_LEFT | B_FOLLOW_TOP_BOTTOM);
	fADC1->SetOrientation(B_VERTICAL);
	fADC1->SetEnabled(false);
	analogBox->AddChild(fADC1);
	fADC1Label = new BStringView(BRect(85, 340, 120, 355), "adc1_label", "0",
		B_FOLLOW_LEFT | B_FOLLOW_BOTTOM);
	fADC1Label->SetAlignment(B_ALIGN_CENTER);
	analogBox->AddChild(fADC1Label);
	// ADC2
	label = new BStringView(BRect(125, 15, 160, 30), "adc2", "ADC2");
	label->SetAlignment(B_ALIGN_CENTER);
	analogBox->AddChild(label);
	fADC2 = new BSlider(BRect(125, 35, 160, 335), "adc2_slider", NULL,
		new BMessage(MSG_SET_OUTPUTS), 0, 255, B_BLOCK_THUMB,
		B_FOLLOW_LEFT | B_FOLLOW_TOP_BOTTOM);
	fADC2->SetOrientation(B_VERTICAL);
	fADC2->SetEnabled(false);
	analogBox->AddChild(fADC2);
	fADC2Label = new BStringView(BRect(125, 340, 160, 355), "adc2_label", "0",
		B_FOLLOW_LEFT | B_FOLLOW_BOTTOM);
	fADC2Label->SetAlignment(B_ALIGN_CENTER);
	analogBox->AddChild(fADC2Label);
	back->AddChild(analogBox);

	// Digital inputs
	inBox = new BBox(BRect(305, 5, 560, 45), "inputs_box");
	inBox->SetLabel("Inputs");
	for (i = 0; i < 5; i++) {
		sprintf(buffer1, "digital_input_%d", i);
		sprintf(buffer2, "%d", i + 1);
		fDigitalIn[i] = new BCheckBox(BRect(5 + i * 30, 15, (i + 1) * 30, 30),
			buffer1, buffer2, NULL);
		fDigitalIn[i]->SetEnabled(false);
		inBox->AddChild(fDigitalIn[i]);
	}
	back->AddChild(inBox);

	// Digital outputs
	outBox = new BBox(BRect(305, 50, 560, 95), "outputs_box");
	outBox->SetLabel("Outputs");
	for (i = 0; i < 8; i++) {
		sprintf(buffer1, "digital_output_%d", i);
		sprintf(buffer2, "%d", i + 1);
		fDigitalOut[i] = new BCheckBox(BRect(5 + i * 30, 15, (i + 1) * 30,
			30), buffer1, buffer2, new BMessage(MSG_SET_OUTPUTS));
		outBox->AddChild(fDigitalOut[i]);
	}
	back->AddChild(outBox);

	// Counters
	fCounter1 = new CounterBox(BRect(305, 100, 430, 200), 1);
	back->AddChild(fCounter1);
	fCounter2 = new CounterBox(BRect(435, 100, 560, 200), 2);
	back->AddChild(fCounter2);

	// Move the window to the center of the screen
	MoveTo((screen.Frame().Width() - Bounds().Width()) / 2,
		(screen.Frame().Height() - Bounds().Height()) / 2);
}

void MyWindow::_update_analog_labels(void)
{
	int dac1 = fDAC1->Value(), dac2 = fDAC2->Value();
	char label[64];
	sprintf(label, "%d", dac1);
	fDAC1Label->SetText(label);
	sprintf(label, "%d", dac2);
	fDAC2Label->SetText(label);
}

void MyWindow::MessageReceived(BMessage *message)
{
	bool handled = false;

	switch (message->what) {

		// Connect/disconnect
		case MSG_CONNECT:
			if (fConnected) {
				CloseDevice();
				fConnect->SetLabel("Connect");
				fConnectLabel->SetText("disconnected");
				fConnected = false;
				fTest = -1;
			}
			else {
				int address = (fSK5->Value() == B_CONTROL_ON ? 0 : 1) +
					(fSK6->Value() == B_CONTROL_ON ? 0 : 2);
				char label[64];
				if (OpenDevice(address) == -1)
					sprintf(label, "Board %d not found", address);
				else {
					sprintf(label, "Connected to board %d", address);
					fConnect->SetLabel("Disconnect");
					fConnected = true;
					fDAC1->SetValue(0);
					fDAC2->SetValue(0);
					_update_analog_labels();
					for (int i = 0; i < 8; i++)
						fDigitalOut[i]->SetValue(B_CONTROL_OFF);
				}
				fConnectLabel->SetText(label);
			}
			break;

		// Buttons
		case MSG_SET_ALL_DIGITAL: {
			SetAllDigital();
			for (int i = 0; i < 8; i++)
				fDigitalOut[i]->SetValue(B_CONTROL_ON);
			} break;
		case MSG_CLEAR_ALL_DIGITAL: {
			ClearAllDigital();
			for (int i = 0; i < 8; i++)
				fDigitalOut[i]->SetValue(B_CONTROL_OFF);
			} break;
		case MSG_SET_ALL_ANALOG:
			SetAllAnalog();
			fDAC1->SetValue(255);
			fDAC2->SetValue(255);
			_update_analog_labels();
			break;
		case MSG_CLEAR_ALL_ANALOG:
			ClearAllAnalog();
			fDAC1->SetValue(0);
			fDAC2->SetValue(0);
			_update_analog_labels();
			break;
		case MSG_OUTPUT_TEST:
			// Test is performed by MyView::Pulse()
			if (fConnected) {
				if (fTest == -1)
					fTest = 0;
				else
					fTest = -1;
			}
			break;

		// Set outputs
		case MSG_SET_OUTPUTS: {
			unsigned char digital = 0x00;
			OutputAllAnalog(fDAC1->Value(), fDAC2->Value());
			_update_analog_labels();
			for (int i = 0; i < 8; i++) {
				if (fDigitalOut[i]->Value() == B_CONTROL_ON)
					digital += (0x01 << i);
			}
			WriteAllDigital(digital);
			} break;

		// Counters
		case MSG_RESET_COUNTER: {
			int counter;
			if (message->FindInt32("counter_number",
			  (int32*)&counter) == B_OK)
				ResetCounter(counter);
			} break;
		case MSG_SET_DEBOUNCE_TIME: {
			int counter, debounce;
			if (message->FindInt32("counter_number",
			  (int32*)&counter) == B_OK &&
			  message->FindInt32("debounce_time", (int32*)&debounce) == B_OK)
				SetCounterDebounceTime(counter, debounce);
			} break;

	}

	if (!handled)
		BWindow::MessageReceived(message);
}

bool MyWindow::QuitRequested(void)
{
	if (fConnected)
		CloseDevice();
	be_app->PostMessage(B_QUIT_REQUESTED);
	return BWindow::QuitRequested();
}

// MyApp ---------------------------------------------------------------------

MyApp::MyApp(void) :
	BApplication("application/bfz.K8055_Demo")
{
	fWindow = new MyWindow();
}

void MyApp::AboutRequested(void)
{
	(new BAlert("About", "Velleman USB Experiment Interface Board K8055 "
		"library for BeOS/ZETA demo\n\n"
		"Version Beta 1 - 2007/02/09\n\n"
		"Copyright (C) 2007 Gabriele Biffi - www.biffuz.it\n\n"
		"Library and application released under the terms of the MIT license",
		"OK", NULL, NULL, B_WIDTH_AS_USUAL, B_INFO_ALERT))->Go();
}

void MyApp::ReadyToRun(void)
{
	fWindow->Show();
}

// main ----------------------------------------------------------------------

int main(int, char**)
{
	(new MyApp())->Run();
	return 0;
}
