/*
Velleman K8055 USB Experiment Interface Board library for BeOS/ZETA
Copyright (c) 2007 Gabriele Biffi
Version Beta 1 - 2007/02/09

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <string.h>

#ifdef K8055_ENABLE_SET_COUNTER_DEBOUNCE_TIME
	#include <math.h>
	#include <stdio.h>
#endif // K8055_ENABLE_SET_COUNTER_DEBOUNCE_TIME

#include <OS.h>

#ifdef _ZETA_
	#include <usb/USBKit.h>
#else
	#include "USBKit.h"
	#define BUSBRoster USBRoster
	#define BUSBDevice USBDevice
#endif // _ZETA_

#include "k8055.h"

// USB board identification. The funny fact of the K8055 is that the USB
// Product ID changes by setting the on-board "address" jumpers! 
#define VELLEMAN_VENDOR_ID		0x10CF
#define K8055_PRODUCT_ID_BASE	0x5500
#define K8055_PRODUCT_ID_LIMIT	0x5503

// K8055 commands
#define K8055_COMMAND_SET_DEBOUNCE_1	0x01
#define K8055_COMMAND_SET_DEBOUNCE_2	0x02
#define K8055_COMMAND_RESET_COUNTER_1	0x03
#define K8055_COMMAND_RESET_COUNTER_2	0x04
#define K8055_COMMAND_SET_OUTPUTS		0x05

// Always useful
typedef unsigned char BYTE;

// Useful macros
#define PACKET_SIZE		8
#define PACKET(x)		BYTE x[PACKET_SIZE]; memset(x, 0, PACKET_SIZE)

// Global vars ---------------------------------------------------------------

class K8055Roster;

// The roster maintains the devices[] array. When a K8055 board is found,
// its BUSBDevice pointer is put in the correct entry of the array,
// according to the "address" set by the board's jumpers SK5 and SK6
K8055Roster *	roster = NULL;
BUSBDevice *	devices[4];

// The currently opened device
BUSBDevice * device = NULL;

// Status of the currently opened device. Every time we want to set an
// output, we must supply the board the status of them all, so we have to
// remember them.
// byte 0 = digital outputs, byte 1 = analog1 output, and byte 2 = analog2
// output
BYTE status[3];

// Read/write and reset functions --------------------------------------------

int ReadK8055(BYTE *packet)
{
	if (device == NULL) return -1;
	ssize_t read = device->ActiveConfiguration()->InterfaceAt(0)->
		EndpointAt(0)->InterruptTransfer(packet, PACKET_SIZE);
	if (read != PACKET_SIZE)
		return -1;
	return PACKET_SIZE;
}

int WriteK8055(BYTE *packet)
{
	if (device == NULL) return -1;
	ssize_t wrote = device->ActiveConfiguration()->InterfaceAt(0)->
		EndpointAt(1)->InterruptTransfer(packet, PACKET_SIZE);
	if (wrote != PACKET_SIZE)
		return -1;
	return PACKET_SIZE;
}

void ResetK8055(void)
{
	if (device == NULL) return;
	status[0] = 0;
	status[1] = 0;
	status[2] = 0;
	ClearAllDigital();
	ResetCounter(1);
	ResetCounter(2);
	SetCounterDebounceTime(1, 2);
	SetCounterDebounceTime(2, 2);
}

// K8055Roster ---------------------------------------------------------------

class K8055Roster : public BUSBRoster
{
public:
	K8055Roster(void) {
		// Clear devices arrays
		devices[0] = NULL;
		devices[1] = NULL;
		devices[2] = NULL;
		devices[3] = NULL;
	};
	virtual status_t K8055Roster::DeviceAdded(BUSBDevice *dev) {
		if (dev->VendorID() == VELLEMAN_VENDOR_ID &&
		  (dev->ProductID() >= K8055_PRODUCT_ID_BASE &&
		  dev->ProductID() <= K8055_PRODUCT_ID_LIMIT)) {
		  	// Get the board "address"
		  	int address = dev->ProductID() - K8055_PRODUCT_ID_BASE;
		  	// Only one card for each address!
		  	if (devices[address] == NULL) {
			  	// Set up the device if needed
			  	if (dev->ActiveConfiguration() == NULL)
			  		dev->SetConfiguration(dev->ConfigurationAt(0));
				//printf("Found device %d\n", address);
				devices[address] = dev;
				return B_OK;
		  	}
		}
		#ifdef _ZETA_
			return BUSBRoster::DeviceAdded(dev);
		#else
			return B_ERROR;
		#endif
	};
	void K8055Roster::DeviceRemoved(BUSBDevice *dev) {
		for (int i = 0; i < 4; i++) {
			if (devices[i] == dev) {
				//printf("Removing device %d", i);
				devices[i] = NULL;
			}
			if (device == dev)
				device = NULL;	// brutal
		}
		#ifdef _ZETA_
			BUSBRoster::DeviceRemoved(dev);
		#endif
	};
};

// Open/close device ---------------------------------------------------------

int OpenDevice(int cardAddress)
{
	// Start the roster
	if (roster == NULL) {
		roster = new K8055Roster();
		roster->Start();
		snooze(1000000);
	}

	// Check card address
	if (cardAddress < 0 || cardAddress > 3)
		return -1;
	if (devices[cardAddress] == NULL)
		return -1;

	// Reset the card
	device = devices[cardAddress];
	ResetK8055();

	return cardAddress;
}

void CloseDevice(void)
{
	ResetK8055();
	device = NULL;
}

// Analog channels -----------------------------------------------------------

int ReadAnalogChannel(int channel)
{
	if (channel != 1 && channel != 2) return -1;
	PACKET(in);
	if (ReadK8055(in) == PACKET_SIZE) {
		if (channel == 1) return in[2];
		if (channel == 2) return in[3];
	}
	return -1;
}

void ReadAllAnalog(int *data1, int *data2)
{
	PACKET(in);
	if (ReadK8055(in) == PACKET_SIZE) {
		if (data1 != NULL) *data1 = in[2];
		if (data2 != NULL) *data2 = in[3];
	}
	else {
		if (data1 != NULL) *data1 = -1;
		if (data2 != NULL) *data2 = -1;
	}
}

void OutputAnalogChannel(int channel, int data)
{
	if (channel == 1)
		OutputAllAnalog(status[1], data);
	else if (channel == 2)
		OutputAllAnalog(data, status[2]);
}

void OutputAllAnalog(int data1, int data2)
{
	PACKET(out);
	status[1] = (BYTE)data1;
	status[2] = (BYTE)data2;
	out[0] = K8055_COMMAND_SET_OUTPUTS;
	out[1] = status[0];
	out[2] = status[1];
	out[3] = status[2];
	WriteK8055(out);
}

void ClearAnalogChannel(int channel)
{
	if (channel == 1)
		OutputAllAnalog(status[1], 0);
	else if (channel == 2)
		OutputAllAnalog(0, status[2]);
}

void ClearAllAnalog(void)
{
	OutputAllAnalog(0, 0);
}

void SetAnalogChannel(int channel)
{
	if (channel == 1)
		OutputAllAnalog(status[1], 255);
	else if (channel == 2)
		OutputAllAnalog(255, status[2]);
}

void SetAllAnalog(void)
{
	OutputAllAnalog(255, 255);
}

// Digital channels ----------------------------------------------------------

void WriteAllDigital(int data)
{
	PACKET(out);
	status[0] = (BYTE)data;;
	out[0] = K8055_COMMAND_SET_OUTPUTS;
	out[1] = status[0];
	out[2] = status[1];
	out[3] = status[2];
	WriteK8055(out);
}

void ClearDigitalChannel(int channel)
{
	if (channel < 1 || channel > 8) return;
	WriteAllDigital(status[0] & ~(0x01 << (channel - 1)));
}

void ClearAllDigital(void)
{
	WriteAllDigital(0x00);
}

void SetDigitalChannel(int channel)
{
	if (channel < 1 || channel > 8) return;
	WriteAllDigital(status[0] | (0x01 << (channel - 1)));
}

void SetAllDigital(void)
{
	WriteAllDigital(0xFF);
}

bool ReadDigitalChannel(int channel)
{
	if (channel < 1 || channel > 8) return false;
	int data = ReadAllDigital();
	if (data == -1) return -1;
	return ((BYTE)data & (0x01 << (channel - 1)));
}

int ReadAllDigital(void)
{
	PACKET(in);
	if (ReadK8055(in) == PACKET_SIZE) {
		// Digital inputs status byte is a bit odd. Inputs are mapped as:
		// MSB  5421---3  LSB
		return (BYTE)(
			(in[0] & 0x10) >> 4 |
			(in[0] & 0x20) >> 4 |
			(in[0] & 0x01) << 2 |
			(in[0] & 0x40) >> 3 |
			(in[0] & 0x80) >> 3);
	}
	return -1;
}

// Counters ------------------------------------------------------------------

void ResetCounter(int counterNumber)
{
	if (counterNumber != 1 && counterNumber != 2) return;
	PACKET(out);
	out[0] = (counterNumber == 1 ? K8055_COMMAND_RESET_COUNTER_1 :
		K8055_COMMAND_RESET_COUNTER_2);
	WriteK8055(out);
}

int ReadCounter(int counterNumber)
{
	if (counterNumber != 1 && counterNumber != 2) return -1;
	PACKET(in);
	if (ReadK8055(in) == PACKET_SIZE) {
		if (counterNumber == 1) return (int)in[5] * 256 + (int)in[4];
		if (counterNumber == 2) return (int)in[7] * 256 + (int)in[6];
	}
	return -1;
}

void SetCounterDebounceTime(int counterNumber, int debounceTime)
{
	if ((counterNumber != 1 && counterNumber != 2) || debounceTime < 0)
		return;

#ifdef K8055_ENABLE_SET_COUNTER_DEBOUNCE_TIME
	double capped;
	PACKET(out);
	BYTE val;

	// This formula is from the libk8055 library for Linux by Sven Lindberg
	capped = (debounceTime > 7450 ? 7450 : debounceTime);
	val = (BYTE)sqrt(capped / 0.115);

	// Found these values on the web
	/*if (debounceTime == 0)
		val = 0;
	else if (debounceTime < 2)
		val = 1;
	else if (debounceTime < 10)
		val = 3;
	else if (debounceTime < 1000)
		val = 8;
	else if (debounceTime < 5000)
		val = 88;
	else
		val = 255;*/

	printf("SetCounterDebounceTime(%d, %d) = %d\n", counterNumber,
		debounceTime, val);

	if (counterNumber == 1) {
		out[0] = K8055_COMMAND_SET_DEBOUNCE_1;
		out[6] = val;
	}
	else {
		out[0] = K8055_COMMAND_SET_DEBOUNCE_2;
		out[7] = val;
	}
#endif
}
